#!/usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
require('source-map-support').install();
process.env.AWS_SDK_LOAD_CONFIG = true;
const updateNotifier = require("update-notifier");
const Errors = require("@tradle/errors");
const errors_1 = require("./errors");
const logger_1 = require("./logger");
const utils_1 = require("./utils");
const pkg = require('../package.json');
const DESC = {
    key: 'key returned by create-data-bundle command'
};
updateNotifier({
    pkg,
    updateCheckInterval: 60 * 60 * 1000
}).notify();
require('dotenv').config({
    path: '.env'
});
const _ = require("lodash");
const utils_2 = require("./utils");
const HELP = `
  Commands:
    validate
    load
    deploy
    create-data-bundle
    create-data-claim

  To get help for a command, use --help, e.g.: tradleconf validate --help
`;
const printHelp = () => logger_1.logger.info(HELP);
const getCommandName = command => {
    if (typeof command === 'string')
        return command;
    return command && command.name();
};
const NODE_FLAGS = [
    'debug',
    'debug-brk'
];
const PROGRAM_OPTS = [
    'local',
    'remote',
    'project'
].concat(NODE_FLAGS);
let matchedCommand;
const normalizeOpts = (...args) => {
    const command = matchedCommand = args.pop();
    const programOpts = _.defaults(_.pick(program, PROGRAM_OPTS), defaults.programOpts);
    if (program.debugBrk) {
        programOpts['debug-brk'] = true;
    }
    const { local, remote } = programOpts;
    if (command.name() !== 'validate') {
        const envType = local ? 'local' : 'remote';
        utils_1.debug(`targeting ${envType} environment`);
    }
    const commandOpts = _.pick(command, command.options.map(o => o.attributeName()));
    commandOpts.args = args;
    return {
        commandOpts,
        programOpts: _.extend(_.omit(programOpts, NODE_FLAGS), {
            lambda: new AWS.Lambda(),
            nodeFlags: _.pick(programOpts, NODE_FLAGS)
        })
    };
};
const createAction = action => (...args) => {
    const { programOpts, commandOpts } = normalizeOpts(...args);
    return run(() => {
        const conf = createConf(programOpts);
        return conf[action](commandOpts);
    });
};
const run = (fn) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    let result;
    try {
        result = yield fn();
    }
    catch (err) {
        process.exitCode = 1;
        if (Errors.matches(err, 'developer')) {
            logger_1.logger.error(err.stack);
        }
        else if (Errors.matches(err, errors_1.Errors.UserAborted)) {
            logger_1.logger.info('command canceled');
        }
        else {
            logger_1.logger.error(err.message);
            (matchedCommand || program).outputHelp();
        }
        return;
    }
    if (result == null) {
        logger_1.logger.info('OK');
    }
    else {
        logger_1.logger.info(utils_2.prettify(result));
    }
});
const program = require('commander');
program
    .version(pkg.version)
    .option('-l, --local', 'target local development environment')
    .option('-r, --remote', 'target remote environment')
    .option('-x, --project [path]', 'path to serverless project on disk')
    .option('--debug', 'invoke serverless function under the debugger')
    .option('--debug-brk', 'invoke serverless function under the debugger');
program.on('--help', printHelp);
if (!process.argv.slice(2).length) {
    program.outputHelp();
}
program.parse(process.argv);
const defaults = {
    programOpts: {
        stackName: process.env.stackName,
        profile: process.env.awsProfile,
        project: process.env.project
    }
};
const { profile = defaults.programOpts.profile } = program;
if (profile) {
    process.env.AWS_PROFILE = profile;
}
const deployCommand = program
    .command('deploy')
    .option('-m, --models', 'deploy models')
    .option('-s, --style', 'deploy style')
    .option('-t, --terms', 'deploy terms')
    .option('-b, --bot', 'deploy bot configuration')
    .option('-a, --all', 'deploy all configuration')
    .option('--dry-run', 'print but don\'t execute')
    .allowUnknownOption(false)
    .action(createAction('deploy'));
const loadCommand = program
    .command('load')
    .option('-m, --models', 'load models')
    .option('-s, --style', 'load style')
    .option('-t, --terms', 'load terms')
    .option('-b, --bot', 'load bot configuration')
    .option('-a, --all', 'load all configuration')
    .option('--dry-run', 'print but don\'t execute')
    .action(createAction('load'));
const validateCommand = program
    .command('validate')
    .option('-m, --models', 'validate models and lenses')
    .option('-s, --style', 'validate style')
    .option('-t, --terms', 'validate terms')
    .option('-b, --bot', 'validate bot configuration')
    .option('-a, --all', 'validate all configuration')
    .allowUnknownOption(false)
    .action(createAction('validate'));
const createDataBundleCommand = program
    .command('create-data-bundle')
    .option('-p, --path <path>', 'path to bundle to create')
    .allowUnknownOption(false)
    .action(createAction('createDataBundle'));
const createDataClaimCommand = program
    .command('create-data-claim')
    .option('-k, --key <key>', DESC.key)
    .option('-c, --claimType <claimType>', '"prefill" or "bulk"')
    .allowUnknownOption(false)
    .action(createAction('createDataClaim'));
const getDataBundleCommand = program
    .command('get-data-bundle')
    .option('-c, --claimId <claimId>', 'claim id returned by create-data-claim command')
    .option('-k, --key <key>', DESC.key)
    .allowUnknownOption(false)
    .action(createAction('getDataBundle'));
const listDataClaimsCommand = program
    .command('list-data-claims')
    .option('-k, --key <key>', DESC.key)
    .allowUnknownOption(false)
    .action(createAction('listDataClaims'));
const initCommand = program
    .command('init')
    .allowUnknownOption(false)
    .action(createAction('init'));
const execCommand = program
    .command('exec <command>')
    .allowUnknownOption(false)
    .action(createAction('exec'));
const invokeCommand = program
    .command('invoke')
    .option('-f, --function-name <functionName>', 'invoke a lambda by name')
    .allowUnknownOption(false)
    .action(createAction('invoke'));
const destroyCommand = program
    .command('destroy')
    .allowUnknownOption(false)
    .action(createAction('destroy'));
const infoCommand = program
    .command('info')
    .allowUnknownOption(false)
    .action(createAction('info'));
const logCommand = program
    .command('log [functionName]')
    .description(`view/tail a function's logs.

Passes options through to awslogs (https://github.com/jorgebastida/awslogs)

Make sure to put spaces between options.

Right: tradleconf log oniotlifecycle -s 1d
Wrong: tradleconf log oniotlifecycle -s1d

`)
    .allowUnknownOption(false)
    .option('-s, --start <time-expression>', 'see awslogs docs')
    .option('-e, --end <time-expression>', 'see awslogs docs')
    .option('-w, --watch', 'tail log')
    .option('-t, --timestamp', 'prints the creation timestamp of each event.')
    .option('-i, --ingestion-time', 'prints the ingestion time of each event.')
    .option('-f, --filter-pattern', 'CloudWatch Logs filter pattern')
    .option('-q, --query', 'CloudWatch Logs query pattern')
    .action(createAction('log'));
const AWS = require('aws-sdk');
const { createConf } = require('./');
program.parse(process.argv);
if (!matchedCommand) {
    throw new Error(`command not found with name: ${process.argv[2]}`);
}
//# sourceMappingURL=cmd.js.map